@@ -247,7 +247,7 @@ class GroupedWindowListApplet extends Applet.Applet {
             settings: {},
             homeDir: GLib.get_home_dir(),
             overlayPreview: null,
-            lastCycled: null,
+            lastCycled: -1,
             lastTitleDisplay: null,
             scrollActive: false,
             thumbnailMenuOpen: false,
@@ -263,7 +263,7 @@ class GroupedWindowListApplet extends Applet.Applet {
             getPanel: () => (this.panel ? this.panel : null),
             getPanelHeight: () => this._panelHeight,
             getPanelIconSize: () => this.getPanelIconSize(),
-            getPanelMonitor: () => Main.layoutManager.monitors[this.panel.monitorIndex],
+            getPanelMonitor: () => this.panel ? Main.layoutManager.monitors[this.panel.monitorIndex] : null,
             getAppSystem: () => Cinnamon.AppSystem.get_default(),
             getAppFromWMClass: (specialApps, metaWindow) => this.getAppFromWMClass(specialApps, metaWindow),
             getTracker: () => this.tracker,
